/*************************************************************************************************

Content List: MVC Adapter
Licensed to users of N2CMS under the terms of the Boost Software License 

Copyright (c) 2013 Benjamin Herila <mailto:ben@herila.net>

Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization obtaining a copy of the 
software and accompanying documentation covered by this license (the "Software") to use, reproduce,
display, distribute, execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to do so, all subject to 
the following:

The copyright notices in the Software and this entire statement, including the above license grant,
this restriction and the following disclaimer, must be included in all copies of the Software, in
whole or in part, and all derivative works of the Software, unless such copies or derivative works 
 are solely in the form of machine-executable object code generated by a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT 
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-
INFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*************************************************************************************************/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using N2.Engine;
using N2.Web.Mvc.Html;
using N2.Web.Parts;

namespace N2.Web.Mvc
{
    [Controls(typeof(ContentList))]
    public class ContentListController
    {
    }

    [Adapts(typeof(ContentList))]
    public class ContentListAdapter : PartsAdapter
    {
        public static string GetHtml(ContentItem model)
        {
            if (model == null)
                return ("{model = null}"); // nothing to do 
            if (!(model is ContentList))
                return ("{adapter failure - Model is not a NewsList}"); // nothing to do 

	        var currentItem = model as ContentList;
            var chH = currentItem.HtmlHeader ?? string.Empty;
            var chF = currentItem.HtmlFooter ?? string.Empty;
            var sb = new System.Text.StringBuilder(50*1024 + chH.Length + chF.Length);
            var allNews = new List<ContentItem>();

			if (currentItem.ContainerLinks == null)
			{
				return @"<div class=""alert alert-error"">Content List: ContainerLinks is null.</div>";
			}
			
			var containerLinks = currentItem.ContainerLinks.Where(c => c != null).ToList();
	        bool any = false;
		    foreach (var containerLink in containerLinks)
		    {
				if (containerLink.LinkedPage == null)
					continue;
			    any = true;

			    var aChildren = containerLink.LinkedPage.GetChildPagesUnfiltered() // get one level of child pages
				    .Where(Content.Is.AccessiblePage())
				    .ToList();

			    if (containerLink.Recursive) // additional levels of child pages -- recurse
			    {
				    allNews.AddRange(
					    aChildren.Concat(aChildren.SelectMany(c => c.GetChildPagesUnfiltered().Where(Content.Is.AccessiblePage())))
						    .Distinct() /* but only keep unique child pages */);
			    }
			    else
			    {
				    allNews.AddRange(aChildren);
			    }
		    }
	        if (!any)
		        sb.Append(
			        @"<div class=""alert alert-warning"">Content List: ContainerLinks is empty. Edit the content list and add at least one content container.</div>");

	        var errorList = currentItem.Exceptions.Select(x => x.ToString(CultureInfo.InvariantCulture)).ToList();

	        if (!String.IsNullOrEmpty(currentItem.Title))
            {
                sb.AppendFormat("<h{0}>{1}</h{0}>", currentItem.TitleLevel, currentItem.Title);
            }

            var newsEnumerable = allNews.Where(Content.Is.AccessiblePage());
            {
                // apply sort order ***
                var sortAscending = currentItem.SortDirection == SortDirection.Ascending;
                switch (currentItem.SortColumn)
                {
                    case ContentSortMode.Expiration:
                        Func<ContentItem, DateTime> expirySortSelector = a => a.Expires != null ? a.Expires.Value : new DateTime();
                        newsEnumerable = sortAscending ? allNews.OrderBy(expirySortSelector) : allNews.OrderByDescending(expirySortSelector);
                        break;
                    case ContentSortMode.Title:
                        Func<ContentItem, String> titleSortSelector = a => a.Title;
                        newsEnumerable = sortAscending ? allNews.OrderBy(titleSortSelector) : allNews.OrderByDescending(titleSortSelector);
                        break;
                    case ContentSortMode.PublishDate:
                        Func<ContentItem, DateTime> dateSortSelector = a => a.Published != null ? a.Published.Value : new DateTime();
                        newsEnumerable = sortAscending ? allNews.OrderBy(dateSortSelector) : allNews.OrderByDescending(dateSortSelector);
                        break;
                }

                // apply filter ***

                if (!currentItem.ShowFutureEvents)
                    newsEnumerable = newsEnumerable.Where(a => a.Published != null && a.Published.Value <= N2.Utility.CurrentTime());

                if (!currentItem.ShowPastEvents)
                    newsEnumerable = newsEnumerable.Where(a => a.Published != null && a.Published.Value >= N2.Utility.CurrentTime());

                if (currentItem.MaxNews > 0)
                    newsEnumerable = newsEnumerable.Take(currentItem.MaxNews);
            }

            if (currentItem.DisplayMode == NewsDisplayMode.HtmlItemTemplate)
                sb.Append(chH);

            DateTime? lastDate = null;
            // why no ForEach here? it turns out that the C# compiler changes the behavior RE: how it deals with foreach
            // variables accessed from within lambda expressions. 
            using (var itemEnumerator = newsEnumerable.GetEnumerator())
                while (itemEnumerator.MoveNext())
                {
                    ContentItem item = itemEnumerator.Current;
                    var itemPublished = item.Published == null ? new DateTime() : item.Published.Value;
                    if (currentItem.SortColumn == ContentSortMode.PublishDate && currentItem.GroupByMonth && (lastDate == null || lastDate.Value.Month != itemPublished.Month))
                    {
                        // new month ***
                        sb.AppendFormat("<h2>{0:MMMM yyyy}</h2>\n", itemPublished);
                        lastDate = item.Published.Value;
                    }

                    var showTitle = !String.IsNullOrEmpty(item.Title);// && item.ShowTitle
                    var text = (item.GetDetail("Text") ?? "").ToString();
                    var summary = (item.GetDetail("Summary") ?? "").ToString();

                    // display either full article or abstract + link ***
                    switch (currentItem.DisplayMode)
                    {
                        case NewsDisplayMode.TitleAndText:
                            if (showTitle)
                                sb.AppendFormat("<h{1}>{0}</h{1}>\n", item.Title ?? "Untitled", currentItem.TitleLevel + 1);
                            Debug.Assert(item.Published != null, "item.Published != null");
                            sb.AppendFormat("<div class=\"date\">{0:MMMM d, yyyy}</div>\n", item.Published.Value);
                            sb.AppendFormat("<div class=\"article\">\n{0}\n</div>\n", text);
                            break;

                        case NewsDisplayMode.TitleAndAbstract:
                            if (!String.IsNullOrEmpty(text))
                            {
                                if (showTitle)
                                    sb.AppendFormat("<h{2}><a href=\"{1}\">{0}</a></h{2}>\n", item.Title ?? "Untitled", item.Url, currentItem.TitleLevel + 1);
                                Debug.Assert(item.Published != null, "item.Published != null");
                                sb.AppendFormat("<div class=\"date\">{0:MMMM d, yyyy}</div>\n", item.Published.Value);
                                sb.AppendFormat("<div class=\"abstract\">\n{0}\n</div>\n", summary);
                                sb.AppendFormat("<a href=\"{0}\">Read more...</a>\n", item.Url);
                            }
                            else
                            {
                                if (showTitle)
                                    sb.AppendFormat("<h{1}>{0}</h{1}>\n", item.Title ?? "Untitled", currentItem.TitleLevel + 1);
                                Debug.Assert(item.Published != null, "item.Published != null");
                                sb.AppendFormat("<div class=\"date\">{0:MMMM d, yyyy}</div>\n", item.Published.Value);
                                sb.AppendFormat("<div class=\"abstract\">\n{0}\n</div>\n", summary);
                            }
                            break;

                        case NewsDisplayMode.TitleLinkOnly:
                            sb.AppendFormat("<h{2}><a href=\"{1}\">{0}</a></h{2}>\n", item.Title ?? "Untitled", item.Url, currentItem.TitleLevel + 1);
                            Debug.Assert(item.Published != null, "item.Published != null");
                            sb.AppendFormat("<div class=\"date\">{0:MMMM d, yyyy}</div>\n", item.Published.Value);
                            break;

                        case NewsDisplayMode.HtmlItemTemplate:
                            try
                            {
                                if (String.IsNullOrWhiteSpace(currentItem.HtmlItemTemplate))
                                {
                                    break;
                                }
                                if (!currentItem.HtmlItemTemplate.Contains("$$") && !currentItem.HtmlItemTemplate.Contains("{{"))
                                {
                                    sb.Append(currentItem.HtmlItemTemplate);
                                    break;
                                }

                                sb.Append(VariableSubstituter.Substitute(currentItem.HtmlItemTemplate, item));
                            }
                            catch (Exception x)
                            {
                                errorList.Add(x.ToString());
                            }
                            break;

                        default:
                            errorList.Add("Can't render web part: invalid NewsDisplayMode.");
                            break;
                    }
                }

            if (currentItem.DisplayMode == NewsDisplayMode.HtmlItemTemplate)
                sb.Append(chF);

            errorList.ForEach(x => sb.AppendFormat(@"<div class=""alert alert-error""><pre>{0}</pre></div>", x));

            return sb.ToString();
        }

        public override void RenderPart(System.Web.Mvc.HtmlHelper html, ContentItem part, System.IO.TextWriter writer = null)
        {
            html.ViewContext.Writer.Write(GetHtml(part));
        }
    }
}
